function [highDimData, samples] = SampleAndProjUp(minX, maxX, lowDimLandmarks, alpha, beta, drMapping)
% The function [highDimData, numSamples1, numSamples2] = CalcVisClassifier(minX, maxX, lowDimLandmarks, alpha, beta)
% samples the 2D space specified by the borders minX and maxX and projects
% these samples up to the high dimensional space.
%
% inputs:
%   minX            : minimal values in both dimensions
%   maxX            : maximal values in both dimensions
%   lowDimLandmarks : low dimensional landmark vectors used for the projection
%   alpha           : parameters of the projection. They live in the
%                     highdimensional space and there should be one per lowDimLandmark
%   beta            : width of the gaussian kernels used in the 2D space
%
% outputs:
%   highDimData     : primary output of this function, the projected
%                     samples from the 2D space to the high dimensional space
%   Samples         : struct containing information about the samples process

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%


stepS = (maxX - minX)./200;

samples1 = minX(1):stepS(1):maxX(1);
samples2 = minX(2):stepS(2):maxX(2);

dataSpace   = sample2Dspace(samples1, samples2);

if nargin >= 6
    switch lower(drMapping.dimRed)
        case {'pca'}
            % inv projections with PCA
            highDimData = bsxfun(@plus, drMapping.mean', drMapping.M * dataSpace);

        case {'som', 'fisher som', 'fisher rsom'}
            % inv projection with SOM
            n2dSamples = size(dataSpace,2);
            nHdDims    = size(drMapping.sMap.codebook,2);

            highDimData = zeros(nHdDims, n2dSamples);
            ldGrid = som_vis_coords(drMapping.sMap.topol.lattice,drMapping.sMap.topol.msize);
            % transform the grid in the same way the projected data have
            % been transformed
            if isfield(drMapping, 'meanLd')
                % subtract only the mean - the rotation is omitted for SOM
                ldGrid = bsxfun(@minus, ldGrid, drMapping.meanLd);
                %ldGrid = bsxfun(@minus, ldGrid, drMapping.meanLd) * drMapping.eigVecs;
            end
            for i=1:n2dSamples
                [~,ind] = min(sum(bsxfun(@minus, dataSpace(:,i)', ldGrid).^2,2));
                highDimData(:,i) = drMapping.sMap.codebook(ind,:)';
            end

        case {'gtm', 'fisher gtm', 'fisher rgtm'}
            FI = gtm_gbf(drMapping.MU, drMapping.sigma, dataSpace');
            highDimDataAlpha = FI * drMapping.W; % these are only the alpha's
            
            highDimData = (highDimDataAlpha * drMapping.highDimData)';
            
            if size(drMapping.highDimData,2) == 3
                % project the prototypes also up and append them to the matrix
                % of the projected samples
                protsAlpha = drMapping.FI * drMapping.W;
                prototypes = (protsAlpha * drMapping.highDimData)';
                highDimData = [highDimData, prototypes];
            end

        case {'nodr', 'no dr'}
            highDimData = dataSpace;
            
        otherwise
            error(['Method for inverting the ' drMapping.dimRed ' mapping is not yet supported.']);
    end
else
    % apply non linear kernel mapping
    sum_ldLandmarks = sum(lowDimLandmarks.^2, 2);
    % project data up
    sum_dataSpace = sum(dataSpace.^2, 1);
    D = bsxfun(@plus, sum_ldLandmarks, bsxfun(@plus, sum_dataSpace, -2*(lowDimLandmarks*dataSpace)));
    %K = bsxfun(@rdivide, D, beta);
    K = bsxfun(@times, D, beta);
    K = exp(-K);
    clear D;

    sum_K = sum(K,1); % normalize, so that each column sums up to one
    K = bsxfun(@rdivide, K, sum_K + 1e-100);


    %fprintf(1, 'Sum_K has the val %f\n', sum_K);

    highDimData = alpha' * K;

    clear K;
end

numSamples1         = length(samples1);
numSamples2         = length(samples2);
samples.samples1    = samples1;
samples.samples2    = samples2;
samples.numSamples1 = numSamples1;
samples.numSamples2 = numSamples2;
samples.dataSpace   = dataSpace;
