function [wNorm, nSVMs] = calcNormW4SVM(model, gamma, debug)
% Given an SVM model and gamma ( = paramter which determines the width of
% the guassian kernel: exp(-gamma*dist)), this method calculates the length
% of the parameters w of each hyperplane.
% Optionaly, it returns also the number of 2 class SVM trainined for the
% multi class problem.

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%


if (nargin == 1)
    gamma = model.Parameters(4);
end

if (nargin < 3)
    debug = 0;
end

% calculate |w| = sqrt(|w|^2) = sqrt(w'w) to get the real distance to the
% hyperplane
nClasses = model.nr_class;
nSVMs    = nClasses*(nClasses-1)/2;
wNorm    = zeros(nSVMs,1);
offset   = 0;
if debug
    fprintf(1, 'Calculating |w| for SVM:\n| ');
end
for iSVM = 1:(nClasses-1)
    % go in this order through the SVMs:
    % 1vs2, 1vs3, ..., 1vsN, 2vs3, 2vs4, ..., 2vsN, ..., N-2vsN-1, N-2vsN, N-1vsN
    % e.g. N=4: 
    % | 1vs2, 1vs3, 1vs4 | 2vs3, 2vs4 | 3vs4 |
    % thats how the rho are saved in the SVM struct
    nSVMsInBlock = nClasses - iSVM;
    for j=1:nSVMsInBlock
        % current SVM is (iSVM)vs(iSVM+j) 
        if debug, fprintf(1,'%dvs%d', iSVM, iSVM+j); end
        if (j ~= nSVMsInBlock)
            if debug, fprintf(1, ', '); end
        end
        % Calc the indices of the SV coefficients
        c1SVs = (sum(model.nSV(1:iSVM-1))+1):sum(model.nSV(1:iSVM));
        c2SVs = (sum(model.nSV(1:iSVM+j-1))+1):sum(model.nSV(1:iSVM+j));
        % find coefficients (=alpha's) and SVs
        coefs = [model.sv_coef(c1SVs, iSVM+j-1); model.sv_coef(c2SVs, iSVM)];
        SVs   = full([model.SVs(c1SVs,:); model.SVs(c2SVs,:)]);
        SVs   = SVs(coefs ~= 0,:);
        coefs = coefs(coefs ~= 0);
        
        sum_SVs = sum(SVs.^2, 2);
        K = bsxfun(@plus, sum_SVs, bsxfun(@plus, sum_SVs', -2*(SVs*SVs')));
        K = exp(-gamma .* K);
        wNorm(offset + j) = coefs'*K*coefs;

        if (nSVMs == 2)
            sv_coeffs = -1 .* sv_coeffs;
        end
    end
    offset = offset + nSVMsInBlock;
    if debug, fprintf(1, ' | '); end
end
if debug, fprintf(1, '\n'); end
 
% % old variant. works only for 2 class problems
% numSVs = size(SVs, 1);
% numC1SVs = model.nSV(1);
% numC2SVs = model.nSV(2);
% L1 = model.Label(1);
% L2 = model.Label(2);
% wNorm_old  = 0;
% for i=1:length(SVs)
%     dist = zeros(numSVs, 1);
%     diff = SVs - repmat(SVs(i,:), numSVs, 1);
%     for j=1:dimX
%         dist = dist + diff(:,j).^2;
%     end
%     dist  = exp(-gamma .* dist);
%     %dist  = dist .* model.sv_coef .* [L1.*ones(numC1SVs,1); L2.*ones(numC2SVs,1)];
%     dist  = dist .* model.sv_coef .* (-1);
%     if (i <= numC2SVs)
%         label = L1;
%     else
%         label = L2;
%     end
%     label = -1;
%     dist  = dist .* model.sv_coef(i) .* label;
%     wNorm_old = wNorm_old + sum(dist);
% end
% wNorm_old = sqrt(wNorm_old)
wNorm = sqrt(wNorm);
