% This demo file demonstrates a few usage scenarios of the Classifier 
% Visualization Toolbox.
% To run all the examples, a few toolboxes are necessary - see the README 
% file.

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%


load toyData.mat


%% --- simple example with the k-nn classifier ---
% ================================================
% we employ here the matlab internal knn method, so no additional
% classifier has to be installed for this first part of the demo

% create a knn classifier
knnModel = fitcknn(X_tr,labels_tr, 'NumNeighbors', 10);

% we have created a function 'predictCertaintyKnn' which uses the standard
% interface to the classifier in order to accept the standardized input /
% output arguments (see also the README file)
% visualize the classifier
invMap = visClassif(X_tr, labels_tr, @predictCertaintyKnn, knnModel, ...
    'dimRed', 'tsne', 'highDimD4Quality', X_te, 'labelsD4Quality', labels_te);

% the function creates 3 plots:
% The first two depict only the certanties (first figure) and only the
% classification (second figure). The last image is a mix of both.
% The first two can be automatically closed by setting the parameter 
% 'plotIntermResult' to 0.

% The evaluation of the visualization is is based on section 3.3 in
% 'Using Discriminative Dimensionality Reduction to Visualize Classifiers',
% http://link.springer.com/article/10.1007%2Fs11063-014-9394-1


%% ------------- example with 3 SVMs -------------
% ================================================
% train three SVMs
% For this demo, the libsvm toolbox is required (see README)
nSVMs = 3;
svmModels = cell(nSVMs,1);
gammas = [0.01, 0.1, 1];
accs = zeros(nSVMs, 2); % first row training accuracies, second test

% examine the effect of different parameters for the model
for i=1:nSVMs
    CSVM         = 10;
    gammaSVM     = gammas(i);
    svmModels{i} = svmtrain(labels_tr, X_tr, ['-c ' sprintf('%f',CSVM) '-q -g ' sprintf('%f',gammaSVM)]);

    % evaluate the performance
    svm_trLabs     = svmpredict(labels_tr, X_tr, svmModels{i}, '-q');
    svm_testLabs   = svmpredict(labels_te, X_te, svmModels{i}, '-q');
    accs(i,1)  = 100*sum(svm_trLabs == labels_tr)/numel(labels_tr);
    accs(i,2)  = 100*sum(svm_testLabs == labels_te)/numel(labels_te);
end

% display the performance of the trained models
fprintf('--------------------------------\n');
fprintf('Accuracies of the SVM models\n');
fprintf('           training     test \n');
for i=1:nSVMs
    fprintf('gamma=%1.2f   %2.1f%%      %2.1f%% \n', gammas(i), accs(i,1), accs(i,2));
end
fprintf('--------------------------------\n');

% Reviewing the classification accuracies, we see that the model improves
% on the training data, but has a stable performance on the test set

%% compute the visualization of the trained classifiers
% first, compute a projection of the data
invMapDR = visClassif(X_tr, labels_tr, @predictCertaintySVM, svmModels{1}, ...
    'dimRed', 'tsne', 'highDimD4Quality', X_te, ...
    'labelsD4Quality', labels_te, 'dr_only', 1);

% Points with an inner black circle depict projections of highDimD4Quality.
% Points which are emphasized with outter circles are positions where the 
% visualization of the classifier is incorrect (considering the label- see 
% section 3.3 in 'Using Discriminative Dimensionality Reduction to
% Visualize Classifiers').
% using an already computed data projection
for i=1:nSVMs
invMap = visClassif(X_tr, labels_tr,  @predictCertaintySVM, svmModels{i}, ...
    'dimRed', 'tsne', 'highDimD4Quality', X_te, ...
    'labelsD4Quality', labels_te, 'invMap', invMapDR, 'plotIntermResult', 0);
end

% to replot an already computed visualization of the classifier, run:
%visClassif(X, Labels, @predictCertaintySVM, svmModels{i}, 'invMap', invMap);

% Now, the reason for the increased training accuracy is evident: the
% class boundaries are more complex for SVM 2 and 3. However, the
% boundaries of SVM 3 are overly complex (particularly for the overlapping
% class region), explaining the slight decrease in test performance.

%% visualize the classifier using a discriminative DR technique
% For a discriminative DR, labels are required. We use the labels of the 
% classifier here, hence this shows the view of the classification model on
% the data.
% Visualize the SVM model number 2
modNum = 2;
svm_trLabs   = svmpredict(labels_tr, X_tr, svmModels{modNum}, '-q');
svm_testLabs = svmpredict(labels_te, X_te, svmModels{modNum}, '-q');

invMapDisc = visClassif(X_tr, svm_trLabs, @predictCertaintySVM, svmModels{modNum}, ...
    'dimRed', 'fisher tsne', 'highDimD4Quality', X_te, ...
    'labelsD4Quality', svm_testLabs, 'nDataForParzen', round(size(X_tr,1)/2), ...
    'plotIntermResult', 0, 'dr_only', 1);

% replot the same visualization using the orignal labels
invMapDisc = visClassif(X_tr, labels_tr, @predictCertaintySVM, svmModels{modNum}, ...
    'dimRed', 'fisher tsne', 'highDimD4Quality', X_te, ...
    'labelsD4Quality', labels_te, 'nDataForParzen', round(size(X_tr,1)/2), ...
    'plotIntermResult', 0, 'invMap', invMapDisc);

% This emphasizes directions where the class distributions change. See e.g.
% the small blue cluster between the two red ones.



%% ------------- simple 3D example -------------
% ==============================================
% 3D toy data from the paper
% 'Using Discriminative Dimensionality Reduction to Visualize Classifiers',
% http://link.springer.com/article/10.1007%2Fs11063-014-9394-1

load 3ClustersAbove;

% train a SVM
CSVM = 10;
gammaSVM = 1;
svmModel = svmtrain(labels, X, ['-c ' sprintf('%f',CSVM) '-q -g ' sprintf('%f',gammaSVM)]);

% visualize the model
invMap = visClassif(X, labels, @predictCertaintySVM, svmModel, ...
    'dimRed', 'tsne', 'plotIntermResult', 1);

% since the data is 3D and 'plotIntermResult' is 1, the method visClassif
% additionally plots the original data and the location of the inverse
% projected samples.




