function A = optAlphaWrtEFish(landmarks, trainData, trainDataTargetDim, beta, A, J, globalLearnRate, verbosity)
% function A = optAlphaWrtEFish(landmarks, trainData, trainDataTargetDim, beta, A, SVMModel, wNorm)
% This method does a gradient descent to optimize the parameters alpha for
% the kernel mapping. It minimizes 
%                sum_i ||y^i - A*k^i||^2_{EFish}
% for A.
%
% parameters (all data should be rows of the matrices):
% landmarks             : given prototypes
% trainData             : data to project (in the same space as landmarks)
% trainDataTargetDim    : targets for trainData in the other space
% beta                  : size of the gaussian kernel
% A                     : heuristicly estimated alpha, used for initialization
% J                     : is a (nDims x nDims x nData) Tensor containing
%                         the Fisher information matrices for each point
% globalLearnRate [opt] : optionaly, specify starting learningrate
%
%
% NOTE: This method is applicable for optimization of both projections 
% (from low to high and high to low dimensional spaces). [not testet yet]

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%


nData = size(trainData, 1);
sum_trData = sum(trainData.^2, 2);
sum_landmarks = sum(landmarks.^2, 2);
D = bsxfun(@plus, sum_landmarks, bsxfun(@plus, sum_trData', -2*(landmarks*trainData')));
K = exp(-bsxfun(@times, D, beta));
K = bsxfun(@rdivide, K, sum(K,1) + 1e-100);


% calc the error before optimization
Yp      = K' * A;
diff    = trainDataTargetDim -  Yp;
errEucl = sum(sum(diff.^2))/size(trainDataTargetDim, 1);

errDisc = 0;
for j=1:size(trainDataTargetDim, 1)
    diff = trainDataTargetDim(j,:) -  Yp(j,:);
    errDisc = errDisc + diff * J(:,:,j) * diff';
end
errDisc = errDisc/size(trainDataTargetDim, 1);
if verbosity
    disp('Optimize the parameters of the inverse mapping via gradient descent on E_Fish');
    fprintf(1,'(euclidean/discriminative) MSE before alpha optim. is %f, %f.\n', errEucl, errDisc);
end

%fprintf(1,'MSE (in euclidean space) before alpha optim. is %f.\n', errEucl);
    


nEpochs  = 40; % 20 % 10 for iwann13 paper
etaStart = 1;
%lambda  = 0.00004; % regularization
%theta   = 0.001; % weighting of the euclidean distance
grad     = 0;


if exist('globalLearnRate', 'var')
    etaStart = globalLearnRate;
end

% over time decaying learning rate
etaEnd = etaStart/15; % / 10
eta    = zeros(1,nEpochs);
for i=0:nEpochs-1
    eta(i+1) = etaStart * (etaEnd/etaStart)^(i/(nEpochs-1));
end

A = A'; % now, alpha_i are columns of A

% Do first a few gradien steps in euclidean error space
for ep=1:5
    ind = randperm(nData);
    for i=1:nData
        % update step to decrease the error in euclidean space
        yp    = A*K(:, ind(i));
        t1    = yp*K(:, ind(i))';
        grad = t1 - trainDataTargetDim(ind(i),:)'*K(:, ind(i))';
        
        % do the gradient step
        A = A - eta(ep) * grad;
        %A = A - eta*(lambda2*grad2 + lambda3*A);
        %A = A - eta*(lambda2*grad2);
        %debug1 = debug1  + abs(grad1);
        %debug2 = debug2  + abs(grad2);
    end
    %sum(debug1(:))
    %sum(debug2(:))

    if verbosity
        Yp  = K' * A';
        errEucl = sum(sum((trainDataTargetDim -  Yp).^2))/size(trainDataTargetDim, 1);
        fprintf(1,'MSE (in euclidean space) in epoch %d is %f.\n', ep, errEucl);
    end
end


% we have normalized the fishermatrix to have diag(J) = 1 (and then reguralized).
% But in euclidean case, diag(J) = nDims, so we should scale the gradient.
eta = eta .* size(trainDataTargetDim,2)/(1+size(trainDataTargetDim,2)*0.1);

% Then do the discriminative update which optimizes the E_{Fisher} function
for ep=1:nEpochs
    %debug1 = zeros(size(A,1), size(A,2));
    %debug2 = zeros(size(A,1), size(A,2));
    ind = randperm(nData);
    for i=1:nData
        % update step to decrease the error in Fisher space
        yp    = A*K(:, ind(i));
        t1    = yp*K(:, ind(i))';
        grad  = t1 - trainDataTargetDim(ind(i),:)'*K(:, ind(i))';
        grad  = J(:,:,ind(i)) * grad;
        
        % do the gradient step
        A = A - eta(ep)*grad;
        %debug1 = debug1  + abs(grad1);
        %debug2 = debug2  + abs(grad2);
    end
    %sum(debug1(:))
    %sum(debug2(:))

    
    if verbosity
        Yp      = K' * A';

        errDisc = 0;
        for j=1:size(trainDataTargetDim, 1)
            diff = trainDataTargetDim(j,:) -  Yp(j,:);
            errDisc = errDisc + diff * J(:,:,j) * diff';
        end
        errDisc = errDisc/size(trainDataTargetDim, 1);
        diff    = trainDataTargetDim -  Yp;
        errEucl = sum(sum(diff.^2))/size(trainDataTargetDim, 1);
        fprintf(1,'(euclidean/discriminative) MSE in epoch %d is %f, %f.\n', ep, errEucl, errDisc);
    end
    % this is slower then the for-loop
%     diff    = trainDataTargetDim -  Yp;
%     errEucl = sum(sum(diff.^2))/size(trainDataTargetDim, 1);
%     diffCell = mat2cell(diff,ones(1,size(diff,1)),size(diff,2));
%     JCell   = mat2cell(J,size(J,1),size(J,2), ones(1,size(J,3)));
%     JCell   = JCell(:);
%     errDisc = cellfun(@(x,y) x * y *x', diffCell, JCell);
%     errDisc = sum(errDisc)/size(trainDataTargetDim, 1);
    
end


A = A';
if verbosity
    fprintf('\n');
end

