function [labels, certainties] = predictCertaintySVM(model, highDimData)
% Calculates the prediction of the SVM. If the model is trained with
% probabilities, then those are returned. Otherwise, distances from the 
% hyperplane are returned.
%
% The inputs:
%   model                - a model trained of the trained classifier
%   highDimData          - the input data to be classified (rows of the
%                          matrix)
%
% The outputs:
%   labels               - the predicted labels for the points in 
%                          highDimData (a column vector)
%   certainties          - values indicating the certainty of the
%                          classification (for the predicted class); should
%                          be a column vector
%
% NOTE: This method is written for the SVM implementation provided by the
%       libsvm toolbox.

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%


wNorm = calcNormW4SVM(model);

nClasses = model.nr_class;

% Returns the labels and a certainty meassure
if isempty(model.ProbA)
    % non probabilistic SVM
    % use evalc to suppress the output of older libsvm versions
    [~, labels, ~, certainties] = evalc('svmpredict(zeros(size(highDimData,1),1), highDimData, model);');
    if nClasses > 2
        certainties = calcDistToNextHypPlane4SVM(model, certainties, labels, wNorm);
    else
        certainties = certainties/wNorm;
    end
else
    % probabilistic SVM
    [labels, ~, certainties] = svmpredict(zeros(length(highDimData),1), highDimData, model, '-b 1');
    
    if nargout > 1
        % the order of labels in sampleLabel is correct, i.e.
        % 1..nClasses, but the order in vals is wrong, i.e.
        % model.Labels - therefore, reorder!
        addrElements = zeros(length(labels),1);
        for i=1:model.nr_class
            addrElements(labels == i) = find(model.Label==i);
        end
        certainties = addressElements(certainties, addrElements);
    end
end
