function [alphaHD, beta, lowDimLandmarks, samples, imRes, finalVals, ...
    finalLabels, classSim] = showClassifCurrentVis(ClassifPtr, model, ...
    lowDimData, highDimData, labels, varargin)
% The function showClassifCurrentVis calculates the visualization of a
% classifier and plots it. It required a 2 dimensional projection of the
% data as well as a trained classifier.
%      
% Necessary parameters are:
%  ClassifPtr       - a function handle executing a prediction of the 
%                     classifier; see 'predictCertaintyMethodName.m' for 
%                     further information
%  model            - the classifier model
%  lowDimData       - low dimensional projections of 'highDimData'
%  highDimData      - the original high dimensinal data
%  labels           - the according labels of 'highDimData'
%  
%
% Optional Parameters are:
%  data2plot        - additional data that should be plotted, e.g. test data
%  labels2plot      - labels of the additional data
%  lowDimLandmarks  - landmarks for the inverse mapping
%  highDimLandmarks - initializations for the parameters alpha of the 
%                     inverse mapping
%  estimateQualy    - evaluate the quality of the visualization
%  showVisErrors    - in addition to the classifier, indicate also the region
%                     where the visualization is inacurate. 1 shows the
%                     train errors and 2 train and test, if provided by 
%                     lowDimD4Quality.
%  highDimD4Quality - high dimensional test data used to estimate the
%                     quality of the visualization of the classifier
%  lowDimD4Quality  - low dimensional test data used to estimate the
%                     quality of the visualization of the classifier
%  J                - a (nDims x nDims x nData) matrix representing the 
%                     Fisher information matrix for each data training
%                     point
%
% NOTE: If an already optimized visualization should be plotted, the
%       following parameters should be provided:
%         - alphaHD
%         - beta
%         - lowDimLandmarks
%
%

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%



p = inputParser;
p.addRequired('ClassifPtr', @(x) isa(x, 'function_handle') || isa(x, 'char'));
p.addRequired('model', @(x) isstruct(x) || isobject(x));
p.addRequired('lowDimData', @(x) (size(x,2)==2) || (size(x,2)==3));
p.addRequired('highDimData', @(x) size(x,1) == size(lowDimData,1));
p.addRequired('labels', @(x) (size(labels,1) == size(lowDimData,1)) || (size(labels,2) == size(lowDimData,1)));

p.addParamValue('lowDimLandmarks', NaN, @(x) size(x,2) <= 3);
p.addParamValue('highDimLandmarks', NaN, @(x) isnan(x) || size(x,2) == size(highDimData,2));
p.addParamValue('data2plot', [], @(x) size(x,2) < 3);
p.addParamValue('labels2plot', [], @isfloat); 
p.addParamValue('data2plotIdx', [], @isfloat); 
p.addParamValue('alphaHD', NaN, @isfloat); 
p.addParamValue('etaAlpha', 1, @(x) numel(x) == 1);
p.addParamValue('beta', NaN, @isfloat);
p.addParamValue('scaleB', 10, @(x) length(x) == 1);
p.addParamValue('lowDimProts', NaN, @isfloat);
p.addParamValue('protLabels', NaN, @isfloat);
p.addParamValue('plotIntermResult', 1, @(x) (x==1) | (x==0));
p.addParamValue('intensityScale', NaN, @(x) numel(x) <= 2);
p.addParamValue('J', NaN, @(x) isnan(x(1,1,1)) || (size(x,1) == size(highDimData,2))); 
p.addParamValue('estimateQualy', 1, @(x) numel(x) == 1);
p.addParamValue('showVisErrors', 1, @(x) numel(x) == 1);
p.addParamValue('highDimD4Quality', [], @(x) isempty(x) || (size(x,2) == size(highDimData,2)));
p.addParamValue('lowDimD4Quality', [], @(x) size(x,2) < 3);
p.addParamValue('drMapping', struct(), @isstruct);
p.addParamValue('verbosity', 0, @(x) (x == 0) || (x == 1));
p.addParamValue('nLandmarks', NaN, @(x) numel(x) == 1);

p.parse(ClassifPtr, model, lowDimData, highDimData, labels, varargin{:});

lowDimLandmarks  = p.Results.lowDimLandmarks;
highDimLandmarks = p.Results.highDimLandmarks;
alphaHD          = p.Results.alphaHD;
beta             = p.Results.beta;
scaleB           = p.Results.scaleB;
plotIntermResult = p.Results.plotIntermResult;
etaAlpha         = p.Results.etaAlpha;
intensityScale   = p.Results.intensityScale;
J                = p.Results.J;
lowDimProts      = p.Results.lowDimProts;
protLabels       = p.Results.protLabels;
data2plot        = p.Results.data2plot;
labels2plot      = p.Results.labels2plot;
data2plotIdx     = p.Results.data2plotIdx;
estimateQualy    = p.Results.estimateQualy;
showVisErrors    = p.Results.showVisErrors;
highDimD4Quality = p.Results.highDimD4Quality;
lowDimD4Quality  = p.Results.lowDimD4Quality;
drMapping        = p.Results.drMapping;
verbosity        = p.Results.verbosity;
nLandmarks       = p.Results.nLandmarks;

if isempty(data2plotIdx)
    data2plotIdx = ones(numel(labels2plot),1);
end

useKernelMapInv = 1;
% if the selected DR technique provides an inverse by itself, don't use the
% kernel mapping
if isfield(drMapping, 'dimRed') 
    if strcmpi(drMapping.dimRed, 'pca') || ...
       strcmpi(drMapping.dimRed(end-2:end), 'som') || ...
       strcmpi(drMapping.dimRed(end-2:end), 'gtm') || ...
       strcmpi(drMapping.dimRed, 'nodr')
        useKernelMapInv = 0;
    end
end
uniqueLabels = unique(labels);
nClasses     = numel(uniqueLabels);

%% generate landmarks if they are not given
if isnan(lowDimLandmarks(1,1)) && useKernelMapInv
    if isnan(nLandmarks)
        nLandmarks = 20*nClasses;
    end
    
    [clusterIdx,lowDimLandmarks] = kmeans(lowDimData, nLandmarks, 'EmptyAction', 'singleton');
    % sometimes, a prototype happens to have coordinates [-Inf -Inf].
    if sum(sum(isinf(lowDimLandmarks))) > 0
        infCoords = find(isinf(lowDimLandmarks));
        lowDimLandmarks(infCoords) = 2*(rand(1, length(infCoords)) -0.5);
    end
    if sum(sum(isnan(lowDimLandmarks))) > 0
        infCoords = find(isnan(lowDimLandmarks));
        lowDimLandmarks(infCoords) = 2*(rand(1, length(infCoords)) -0.5);
    end

    %plot(lowDimLandmarks(:, 1), lowDimLandmarks(:,2), 'o','markeredgecolor','k','markerfacecolor','w','markersize',4);

    % find the nearest data point to each landmark in order to init the
    % highDimLandmarks
    nnIdx = zeros(nLandmarks,1);
    for i=1:nLandmarks
        cIdx = clusterIdx==i;
        % if cluster i contains points, then pick the closest of that
        % cluster
        if sum(cIdx) ~= 0
            [~,elIdx] = min(sum(bsxfun(@minus, lowDimData(cIdx,:), lowDimLandmarks(i,:)).^2,2));
            onesIdx   = find(cIdx);
            nnIdx(i)  = onesIdx(elIdx);
        else
            % if this cluster contains no points, search among all
            % points for the closest
            [~,elIdx] = min(sum(bsxfun(@minus, lowDimData, lowDimLandmarks(i,:)).^2,2));
            nnIdx(i)  = elIdx;
        end
    end
    highDimLandmarks = highDimData(nnIdx,:);
end

%% set the parameters for the inv projection: alpha and beta
if isnan(beta(1))
    sum_ldLandmarks = sum(lowDimLandmarks.^2, 2);

    D = bsxfun(@plus, sum_ldLandmarks, bsxfun(@plus, sum_ldLandmarks', -2*(lowDimLandmarks*lowDimLandmarks')));
    D(1:size(D,1)+1:end) = inf;
    beta = min(D,[],2) .* scaleB;
    beta = mean(beta); % ------- use scalar beta
    %for ilabel=1:numClasses
        % choose beta as mean for each class
        %beta(newc_X==ilabel) = mean(beta(newc_X==ilabel));
    %end
    % D(1:size(D,1)+1:end) = 0; D is not used from here on
    beta = 1./beta;
end

if isnan(alphaHD(1,1)) && useKernelMapInv
    if isnan(J(1,1,1))
        sig_dt = 1.06*size(highDimData,1)^(-1/5);
        % normalize the data before computing fisher matrices
        disp('Calculate the Fisher distances using the parzen window estimator...');
        highDimData_n = bsxfun(@minus, highDimData, mean(highDimData));
        highDimData_n = bsxfun(@rdivide, highDimData_n, std(highDimData_n));
        Dis = pdist2(highDimData_n, highDimData_n);
        J   = estimate_parzen(highDimData_n, labels, Dis, sig_dt);
        clear highDimData_n;
    end
    alphaHD = optAlphaWrtEFish(lowDimLandmarks, lowDimData, highDimData, ...
        beta, highDimLandmarks, J, etaAlpha, verbosity);
    
    % unsupervised inverse projection
    %alphaHD = updateAlphaBatch(lowDimLandmarks, lowDimData, highDimData, beta, highDimLandmarks);
    %alphaHD = highDimLandmarks;
end

%% Sample the 2D space and project the samples up
range([1 3]) = min(lowDimData);
range([2 4]) = max(lowDimData);


minX = [range(1) range(3)].*1.7; % 1.4 previously
maxX = [range(2) range(4)].*1.7;

if useKernelMapInv
    [highDimSamples, samples] = SampleAndProjUp(minX, maxX, lowDimLandmarks, alphaHD, beta);
else
    [highDimSamples, samples] = SampleAndProjUp(minX, maxX, lowDimLandmarks, alphaHD, beta, drMapping);
    if isfield(drMapping, 'dimRed') 
        if strcmpi(drMapping.dimRed(end-2:end), 'gtm') && size(highDimData, 2) == 3
            prototypes = highDimSamples(:,length(samples.dataSpace)+1:end)';
            highDimSamples = highDimSamples(:,1:length(samples.dataSpace));
        end
    end
end
numSamples1   = samples.numSamples1;
numSamples2   = samples.numSamples2;

%% plot the highDimSamples into the 3D data
% in case highDimData are 3-dimensional
if (size(highDimData,2) == 3) && plotIntermResult
    % plot the samples ...
%     myFigure; scatter3(highDimData(:,1), highDimData(:,2), ...
%        highDimData(:,3), [], labels, 'filled'); hold on;
%     plot3(highDimSamples(1,:)', highDimSamples(2,:)', highDimSamples(3,:)', '.g');
    
    % ... or plot a surf
    myFigure; hold on; colorVals = hsv(nClasses);
    if nClasses==2
        colorVals = [1 0 0; 0 0 1];
    end
    for i=1:nClasses
        plot3(highDimData(labels==uniqueLabels(i),1), ... 
                highDimData(labels==uniqueLabels(i),2), ...
                highDimData(labels==uniqueLabels(i),3), 'o', ...
                'MarkerFaceColor', colorVals(i,:), 'Color', colorVals(i,:));
    end
    
    % reorder samples
    xS = zeros(numSamples1, numSamples2);
    for i=1:numSamples2
        xS(:, i) = highDimSamples(1,numSamples1*(i-1) + 1:numSamples1*i)';
    end
    yS = zeros(numSamples1, numSamples2);
    for i=1:numSamples2
        yS(:, i) = highDimSamples(2,numSamples1*(i-1) + 1:numSamples1*i)';
    end
    zS = zeros(numSamples1, numSamples2);
    for i=1:numSamples2
        zS(:, i) = highDimSamples(3,numSamples1*(i-1) + 1:numSamples1*i)';
    end
    % subsample the inverse projected points, so the plot doesn't lag...
    stepS4S = 7;
    surf(xS(1:stepS4S:end,1:stepS4S:end), yS(1:stepS4S:end,1:stepS4S:end), ...
        zS(1:stepS4S:end,1:stepS4S:end));
    colormap(winter); % or bone
    
    
    % plot the grid for the GTM
    if isfield(drMapping, 'dimRed') 
        if strcmpi(drMapping.dimRed(end-2:end), 'gtm')
            plotData(highDimData, labels);
            nSqrtProts = sqrt(size(drMapping.X_prot,1));
            xS = zeros(nSqrtProts,nSqrtProts);
            for i=1:nSqrtProts
                xS(:, i) = prototypes(nSqrtProts * (i-1) +1:nSqrtProts*i, 1);
            end
            yS = zeros(nSqrtProts,nSqrtProts);
            for i=1:nSqrtProts
                yS(:, i) = prototypes(nSqrtProts * (i-1) +1:nSqrtProts*i, 2);
            end
            zS = zeros(nSqrtProts,nSqrtProts);
            for i=1:nSqrtProts
                zS(:, i) = prototypes(nSqrtProts * (i-1) +1:nSqrtProts*i, 3);
            end
            surf(xS, yS, zS);
        end
    end
    
    alpha(.8);
    axis equal;
end

%% classify the projected data
if isa(ClassifPtr, 'function_handle')
    [sampleLabel, vals] = ClassifPtr(model, highDimSamples');
else
    error('classifPtr should be a pointer to a function!');
end

% rearrange the vals
finalVals = zeros(numSamples1, numSamples2);
for i=1:numSamples2
    finalVals(:, i) = vals(numSamples1*(i-1) + 1:numSamples1*i);
end
finalVals = -1*abs(finalVals);

% rearrange labels as well
finalLabels = zeros(numSamples1, numSamples2);
for i=1:numSamples2
    finalLabels(:, i) = sampleLabel(numSamples1*(i-1) + 1:numSamples1*i);
end

%min(finalVals(:))
finalVals = finalVals./abs(min(finalVals(:)));
%intensityScale = -1;

samples.sampleLabel = sampleLabel;

%% estimate the quality of the visualization
format shortG; % for nice display
visErrors = [];
classSim  = zeros(2,1);
% Estimate the quality of the visualization by counting the points which
% are classified differently by the original classifier and the
% lowdimensional visualized classifier.
% Additionaly, save the points for which this is true.
if estimateQualy
    disp('________________________________________________________________');
    % Train points are used here.
    lowDimSamples = samples.dataSpace;
    sampleLabel   = samples.sampleLabel;
    nearestSample = zeros(1, length(lowDimData));
    % Do the classification in the low dim space by comparison to the
    % sampled pixels.
    % This can be done efficent with binary search because points in 
    % dataSpace are sorted. But it is done slowly, here.
    for i=1:length(lowDimData)
        [~, nearestSample(i)] = min(sum(bsxfun(@minus, lowDimSamples, lowDimData(i,:)').^2));
    end
    projL = sampleLabel(nearestSample);
    rx    = vals(nearestSample);
    
    if isa(ClassifPtr, 'function_handle')
        [LabelsClassif, rxOrig] = ClassifPtr(model, highDimData);
        
        % calc the accordance of the labels
        acc         = (LabelsClassif == projL);
        classSim(1) = sum(acc)/length(projL) * 100;
        SimStr      = sprintf('%2.3f', classSim(1));
        disp(['-> ' SimStr '% of the train data were classified with the same label.']);
        
        % calc also the accordance of r(x), the classifier certainties
        pearsonCorr  = corr(rx, rxOrig, 'type', 'Pearson');
        spearmanCorr = corr(rx, rxOrig, 'type', 'Spearman');
        corrs        = [pearsonCorr, spearmanCorr];
        fprintf(['-> Pearson and Spearman correlation values for the classifier \n' ...
            'certainties on train data are %1.3f and %1.3f. \n'], corrs(1), corrs(2));
    
        if (showVisErrors > 0)
            visErrors = [visErrors; lowDimData(logical(1-acc),:)];
        end
    else
        warning('A function handle is required for the evaluation of the visualization!');
    end
    
    
    % Test points are used here.
    if ~isempty(highDimD4Quality) && ~isempty(lowDimD4Quality)
        nearestSample = zeros(1, length(lowDimD4Quality));
        for i=1:length(lowDimD4Quality)
            [~, nearestSample(i)] = min(sum(bsxfun(@minus, lowDimSamples, lowDimD4Quality(i,:)').^2));
        end
        projL = sampleLabel(nearestSample);
        rx    = vals(nearestSample);
        
        if isa(ClassifPtr, 'function_handle')
            [LabelsClassif, rxOrig] = ClassifPtr(model, highDimD4Quality);
            
            % calc the accordance of the labels
            acc         = (LabelsClassif == projL);
            classSim(2) = sum(acc)/length(projL) * 100;
            SimStr      = sprintf('%2.3f', classSim(2));
            disp(['-> ' SimStr '% of the test data were classified with the same label.']);
            
            % calc also the accordance of r(x), the classifier certainties
            pearsonCorr  = corr(rx, rxOrig, 'type', 'Pearson');
            spearmanCorr = corr(rx, rxOrig, 'type', 'Spearman');
            corrs        = [pearsonCorr, spearmanCorr];
            fprintf(['-> Pearson and Spearman correlation values for the classifier \n' ...
                'certainties on test data are %1.3f and %1.3f. \n'], corrs(1), corrs(2));

            if (showVisErrors > 1)
                visErrors = [visErrors; lowDimD4Quality(logical(1-acc),:)];
            end
        else
            warning('A function handle is required for the evaluation of the visualization!');
        end
        
        
    end
    disp('________________________________________________________________');
end

%% estimate the quality by comparing the orig r(x) to the visualized one
% works only for smooth decision boundaries

estimateQual2 = 0;
if estimateQual2
    lowDimSamples = samples.dataSpace;
    vals = abs(vals);
    % identift all samples belonging to the class border
    decBound = lowDimSamples(:, vals < 0.01)';
    
    % compute the distance to the boundary in HD
    [~, rxOrig]  = ClassifPtr(model, highDimData); % simple way
    %listOfOtherClassN = zeros(size(lowDimData,1), 1);
    
    
    
    % decision from the class boundary in the projection space
    rx = zeros(size(highDimData,1), 1);
    
    % determine for each data point the eucl. distance to the dec. boundary
    for i = 1:size(lowDimData,1)
        currPoint = lowDimData(i,:);
        rx(i) = sqrt(min(sum(bsxfun(@minus, decBound, currPoint).^2,2)));
    end
    
    % correlate the two distances
    rxOrig       = abs(rxOrig);
    pearsonCorr  = corr(rx, rxOrig, 'type', 'Pearson');
    spearmanCorr = corr(rx, rxOrig, 'type', 'Spearman');
    corrs        = [pearsonCorr, spearmanCorr];
    fprintf(['Pearson and Spearman correlation values for the distance \n' ...
        'to the class boundary are %1.3f and %1.3f. \n'], corrs(1), corrs(2));
end

%% plot the classifier
imRes = nan;
if (plotIntermResult || nargout >= 4 )
    imRes = plotClassif(lowDimData, labels, samples, finalVals, finalLabels, ...
        'hsv', intensityScale, plotIntermResult, lowDimProts, protLabels, ...
        data2plot, labels2plot, data2plotIdx, visErrors);
end




