function alpha = updateAlphaBatch(landmarks, trainData, trainDataTargetDim, beta, A, numSingVs)
% function alpha = updateAlpha(landmarks, trainData, trainDataTargetDim, beta, A)
% This method does a batch calculation of the parameters alpha for the
% kernel mapping. It solves the equation Y = A*K for A with the pseudo
% inverse.
% This method is applicable for optimization of both projections (from low 
% to high and high to low dimensional spaces). 
%
% parameters (all data should be rows of the matrices):
% landmarks          : given prototypes
% trainData          : data to project (in the same space as landmarks)
% trainDataTargetDim : targets for trainData in the other space
% beta               : size of the gaussian kernel
% A [opt.]           : heuristicly estimated alpha, used for evaluation
% num SingVs [opt.]  : specifies how many singular values to use from the SVD

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%


cmpError = 0;
if exist('A', 'var')
    if ~isempty(A)
        cmpError = 1;
    end
end

sum_trData = sum(trainData.^2, 2);
sum_landmarks = sum(landmarks.^2, 2);
D = bsxfun(@plus, sum_landmarks, bsxfun(@plus, sum_trData', -2*(landmarks*trainData')));
K = exp(-bsxfun(@times, D, beta));
K = bsxfun(@rdivide, K, sum(K,1) + 1e-8);


if (cmpError)
    % calc the error before optimization
    Yt  = K' * A;
    err = sum(sum((trainDataTargetDim -  Yt).^2))/size(trainDataTargetDim, 1);
    fprintf(1,'MSE before alpha optim. is %f.\n', err);
end

% thresh = 5e-2;
% alpha = pinv(K', thresh) * trainDataTargetDim;s
% %alpha = pinv(K') * trainDataTargetDim;
% 
% [U,S,V] = svd(K',0);
% l = min(size(K));
% s = diag(S);
% s = s(1:l); % use only the values according to the smaller dim of K
% size(s)
% %s'
% min(s)

%sum(s > thresh)


% my own pseudo inverse via singular value decomposition
% now, the number of singular values to use can be specified directly
% for using all set:
%numSingVals = size(K,1);
%numSingVals = 40;
numSingVals = rank(K);

if exist('numSingVs', 'var')
    if (numSingVs <= numSingVals)
        % safety heuristic: use only the user specified value if its
        % smaller then the matlab intern heuristic
        numSingVals = numSingVs;
    else
        warning(['Kernel Matrix inversion: Specified value for the number of ' ...
            'singular values (' int2str(numSingVs) ') is larger then the ' ... 
            'estimated rank of the Matrix (' int2str(numSingVals) '). ' ...
            'Using the estimated rank instead.']);
    end
else
    numSingVals = round(numSingVals/2.5); %5
end

[U,S,V] = svd(K', 0);
s       = diag(S);
% disp the smalles current SV
disp(['The smalles current Singular Value is ' sprintf('%s', s(numSingVals)) '.']);
s       = diag(1./s(1:numSingVals));
KtrInv  = V(:,1:numSingVals) * s * U(:,1:numSingVals)'; % K transpose, Inv
alpha   = KtrInv * trainDataTargetDim;

% alternativly, use std regularization
%alpha   = trainDataTargetDim' * K'*(K*K' + 0.05.*eye(size(K, 1)))^(-1);
%alpha   = alpha';

overallNumSVs = size(K,1);
disp(['Using the first ' int2str(numSingVals) ' singular values out of ' int2str(overallNumSVs) ' for the estimation of alpha.']);

if (cmpError)
    % calc the error after optimization
    Yt  = K' * alpha;
    err = sum(sum((trainDataTargetDim -  Yt).^2))/size(trainDataTargetDim, 1);
    fprintf(1,'MSE after alpha optim. is %f.\n', err);
end
fprintf('\n');

