function [invMap, imRes, classSim] = visClassif(highDimData, labels, classifPtr, model, varargin)
% The function visClassif computes a visualization of a given classifier
% using dimensionality reduction. It is based on a general concept and, 
% hence, is applicable to many classification models. See the 'README' file 
% for a detailed explanation or the 'demo' file for example usage.
%      
% Necessary parameters are:
%  highDimData      - the original high dimensional data
%  labels           - the according labels of 'highDimData'
%  ClassifPtr       - a function handle executing a prediction of the 
%                     classifier; see 'predictCertaintyMethodName.m' for 
%                     further information
%  model            - the classifier model
%
% The outputs are:
%  invMap           - struct containing the low-dimensional projections as
%                     well information concerning the classifier
%                     visualization
%  imRes            - the resulting visualization as an image
%  classSim         - the accuracy of the visualization of the classifier
%
%
% Optional Parameters are:
%  dimRed           - dimensionality reduction (DR) used for the 
%                     visualization; typical choices are: 't-SNE', 'pca',
%                     'isomap', 'fisher t-SNE', 'fisher mds'.
%  drParam          - parameter for the dimensionality reduction, depending
%                     on the chosen DR technique. This is the perplexity
%                     for t-SNE, the number of neighbors for Isomap and
%                     [gridSize1, gridSize2, trainLen] for the SOM.
%  invMap           - struct containing already computed information (it is
%                     returned by this method) in order to plot an already
%                     optimized visualization
%  data2plot        - additional data that should be plotted, e.g. test data
%  labels2plot      - labels of the additional data
%  data2plotIdx     - This vector allos to plot several sets of data.
%                     Hence, this vector should indicate to which set each
%                     datapoint in data2plot belongs, e.g. ones for set 1
%                     and twos for set 2.
%  lowDimData       - already computed projection of the data; numData x 2
%  estimateQualy    - evaluate the quality of the visualization
%  showVisErrors    - in addition to the classifier, indicate also the 
%                     regions where the visualization is inacurate. 1 shows 
%                     the train errors and 2 train and test, if  
%                     'lowDimD4Quality' is available.
%  highDimD4Quality - high dimensional test data used to estimate the
%                     quality of the visualization of the classifier
%  lowDimD4Quality  - low dimensional test data used to estimate the
%                     quality of the visualization of the classifier
%  labelsD4Quality  - if labels are given, 'lowDimD4Quality' are also
%                     plotted (if the latter are not set, they will be 
%                     computed  from 'highDimD4Quality')
%  J                - a (nDims x nDims x nData) matrix representing the 
%                     Fisher information matrix for each data training
%                     point
%  dis_Fisher       - fisher distances between the data
%  nDataForParzen   - number of data points that should be utilized to
%                     estimate the parzen window model for 'J'
%  plotIntermResult - boolean indicating whether two intermediate plots
%                     should be displayed (1) or closed (0)
%  dr_only          - boolean indicating whether only a dimensionality
%                     reduction (1) or not (0) should be performed.
%  protsName        - in case of a prototype based classifier, the name of
%                     the field under which the prototypes are saved in the
%                     model struct (if they should be visualized as well).
%                     should be numPrototypes x numDims
%  protsLabelsName  - in case of a prototype based classifier, the labels
%                     of the prototypes (for visualization)
%  alphaHD          - parameters which are optimized for the inverse
%                     mapping
%  betaHD           - gaussian width for the inverse mapping
%  scaleB           - scaling parameter for 'betaHD'
%  lowDimLandmarks  - landmarks for the inverse mapping
%  nLandmarks       - number of landmarks used for the inverse mapping
%  highDimLandmarks - initializations for the parameters 'alphaHD' of the 
%                     inverse mapping
%  etaAlpha         - global learning rate for optimizing of 'alphaHD'
%  intensityScale   - set the limits of the color scale for the decision 
%                     boundary plot manually. Should be [-higher limit,
%                     -lower limit]. If only a single number is given it
%                     is assumed to be the higher limit.
%  verbosity        - the boolean verbosity controls how much debug output
%                     the toolbox will produce (0 implies little, 1 much). 
%
%
% NOTE: If an already optimized visualization should be plotted, the
%       struct invMap should be provided (or alternatively the required 
%       parameters individually).

%
% Copyright (c) 2015 Alexander Schulz
%
% This file is part of the Classifier Visualization Toolbox.
%
% The Classifier Visualization Toolbox is distributed under the
% GNU General Public License (version 3 or later). For details, see
% <http://www.gnu.org/licenses/>.
%


% parse Input
p = inputParser;
p.addRequired('highDimData', @(x) isfloat(x));
p.addRequired('labels', @(x) length(x) == size(highDimData,1));
p.addRequired('ClassifPtr', @(x) isa(x, 'function_handle') || isa(x, 'char'));
p.addRequired('model', @(x) isstruct(x) || isobject(x));

p.addParamValue('dimRed', 't-SNE', @ischar);
p.addParamValue('alphaHD', NaN, @isfloat);
p.addParamValue('betaHD', NaN, @isfloat);
p.addParamValue('scaleB', 10, @(x) length(x) == 1);
p.addParamValue('nLandmarks', NaN, @(x) numel(x) == 1);
p.addParamValue('lowDimData', NaN, @(x) size(x,1) == size(highDimData,1));
p.addParamValue('invMap', struct, @(x) isstruct(x));
p.addParamValue('protsName', '', @ischar);
p.addParamValue('protsLabelsName', '', @ischar);
p.addParamValue('etaAlpha', 1, @(x) numel(x) == 1);
p.addParamValue('intensityScale', NaN, @(x) numel(x) <= 2); 
p.addParamValue('plotIntermResult', 1, @(x) (x==0) || (x==1));
p.addParamValue('data2plot', [], @(x) size(x,2) < 3);
p.addParamValue('labels2plot', [], @isfloat); 
p.addParamValue('data2plotIdx', [], @isfloat);
p.addParamValue('estimateQualy', 1, @(x) numel(x) == 1);
p.addParamValue('showVisErrors', 1, @(x) numel(x) == 1);
p.addParamValue('highDimD4Quality', [], @(x) isempty(x) || (size(x,2) == size(highDimData,2)));
p.addParamValue('lowDimD4Quality', [], @(x) size(x,2) < 3);
p.addParamValue('labelsD4Quality', [], @(x) isfloat(x));
p.addParamValue('lowDimLandmarks', NaN, @(x) size(x,2) < 3);
p.addParamValue('highDimLandmarks', NaN, @(x) size(x,2) == size(highDimData,2));
p.addParamValue('dis_Fisher', NaN, @(x) size(x,1) == size(x,2));
p.addParamValue('J', NaN, @(x) isnan(x(1,1,1)) || size(x,1) == size(x,2));
p.addParamValue('nDataForParzen', 200, @(x) numel(x) == 1);
p.addParamValue('drParam', NaN, @(x) isfloat(x));
p.addParamValue('dr_only', 0, @(x) (x == 0) || (x == 1));
p.addParamValue('verbosity', 0, @(x) (x == 0) || (x == 1));


p.parse(highDimData, labels, classifPtr, model, varargin{:});

alphaHD          = p.Results.alphaHD;
betaHD           = p.Results.betaHD;
scaleB           = p.Results.scaleB;
nLandmarks       = p.Results.nLandmarks;
lowDimData       = p.Results.lowDimData;
dimRed           = p.Results.dimRed;
invMap           = p.Results.invMap;
protsName        = p.Results.protsName;
protsLabelsName  = p.Results.protsLabelsName;
etaAlpha         = p.Results.etaAlpha;
intensityScale   = p.Results.intensityScale;
plotIntermResult = p.Results.plotIntermResult;
data2plot        = p.Results.data2plot;
labels2plot      = p.Results.labels2plot;
data2plotIdx     = p.Results.data2plotIdx;
estimateQualy    = p.Results.estimateQualy;
showVisErrors    = p.Results.showVisErrors;
highDimD4Quality = p.Results.highDimD4Quality;
lowDimD4Quality  = p.Results.lowDimD4Quality;
labelsD4Quality  = p.Results.labelsD4Quality; % todo
highDimLandmarks = p.Results.highDimLandmarks;
lowDimLandmarks  = p.Results.lowDimLandmarks;
dis_Fisher       = p.Results.dis_Fisher;
J                = p.Results.J;
nDataForParzen   = p.Results.nDataForParzen;
drParam          = p.Results.drParam;
dr_only          = p.Results.dr_only;
verbosity        = p.Results.verbosity;


drMapping = struct();
classSim  = nan;

if isfield(invMap, 'alphaHD')
    alphaHD = invMap.alphaHD;
end
if isfield(invMap, 'betaHD')
    betaHD = invMap.betaHD;
end
if isfield(invMap, 'lowDimLandmarks')
    lowDimLandmarks = invMap.lowDimLandmarks;
end
if isfield(invMap, 'J')
    J = invMap.J;
end
if isfield(invMap, 'lowDimProts')
    lowDimProts = invMap.lowDimProts;
else
    lowDimProts = NaN;
end
if isfield(invMap, 'lowDimData')
    lowDimData = invMap.lowDimData;
end
if isfield(invMap, 'drMapping')
    drMapping = invMap.drMapping;
end
if isfield(invMap, 'lowDimD4Quality')
    lowDimD4Quality = invMap.lowDimD4Quality;
end

if isfield(model, protsLabelsName)
    protLabels = model.(protsLabelsName);
else
    protLabels = NaN;
end


% shift the labels
shift  = (min(labels)-1);
labels = labels - shift;

%% calculate the fisher metric
if length(dimRed) >= 6
    if  isnan(lowDimData(1,1)) && strcmpi(dimRed(1:6), 'fisher') && isnan(dis_Fisher(1,1)) %isnan(J(1,1,1))
        if isempty(lowDimD4Quality)
            highDimData = [highDimD4Quality; highDimData];
            labels      = [labelsD4Quality - shift; labels];
        end
        % if the method is prototype based, project also the prototypes
        if ~isempty(protsName)
            highDimData = [model.(protsName); highDimData];
            labels      = [model.(protsLabelsName) - shift; labels];
            % matlab prefers this over getfield(model, protsName).
        end
        
        if nDataForParzen > size(highDimData, 1)
            nDataForParzen = size(highDimData, 1);
        end
        
        % use the parzen window estimator
        disp('Calculate the Fisher distances using the parzen window estimator...');
        highDimData_n = bsxfun(@minus, highDimData, mean(highDimData));
        highDimData_n = bsxfun(@rdivide, highDimData_n, std(highDimData_n));

        % use only a part of the data  for the parzen window estimator
        %[~,~, idxFisher{1}, idxFisher{2}] = subSample(highDimData, labels, nDataForParzen);
        idxFisher = generate_indices(labels, nDataForParzen/size(highDimData, 1));
        [dis_Fisher, J] = compute_fisher_distance(...
            highDimData_n(idxFisher{1},:), labels(idxFisher{1}), highDimData_n, highDimData_n);
        
        
        if isempty(lowDimD4Quality)
            highDimData = highDimData(size(highDimD4Quality,1)+1:end,:);
            labels      = labels(size(highDimD4Quality,1)+1:end);
        end
        if ~isempty(protsName)
            nProts      = size(model.(protsName),1);
            highDimData = highDimData(nProts+1:end,:);
            labels      = labels(nProts+1:end);
        end
        
        dis_Fisher = 0.5 *(dis_Fisher + dis_Fisher');
        disp('... finished calculating the Fisher distances.');
    end
end

%% project the high dimensional data down to 2D
if isnan(lowDimData(1,1)) || (isnan(lowDimProts(1,1)) && ~isempty(protsName))
    if isempty(lowDimD4Quality)
        highDimData = [highDimD4Quality; highDimData];
    end
    % if the method is prototype based, project also the prototypes
    if ~isempty(protsName)
        highDimData = [model.(protsName); highDimData];
        % matlab prefers this over getfield(model, protsName).
    end
    switch lower(dimRed)
        case {'nodr', 'no dr'}
            lowDimData = highDimData;
            
        case {'pca'}
            [lowDimData, drMapping] = pca(highDimData, 2);
            
        case {'fisher mds'}
            lowDimData = cmdscale(dis_Fisher, 2);

        case {'tsne', 't-sne'}
            preProcessDim = 50;
            perpl = 30;
            if ~isnan(drParam)
                perpl = drParam;
            end
            if size(highDimData, 2) < preProcessDim
                preProcessDim = size(highDimData, 2);
            end
            if verbosity
                lowDimData = tsne(highDimData, [], 2, preProcessDim, perpl);
            else
                [~, lowDimData] = evalc('tsne(highDimData, [], 2, preProcessDim, perpl);');
            end
            
            drMapping.perpl         = perpl;
            drMapping.preProcessDim = preProcessDim;
            
        case {'b-h-tsne', 'bh-tsne', 'bhtsne', 'barnes-hut-sne'}
            preProcessDim = 30;
            perpl = 50;
            % degree of approximation: 0 is exact and slow, 1 yields crude
            % approximations
            approxDeg = 0.5;
            if ~isnan(drParam)
                perpl = drParam(1);
                if numel(drParam) > 1
                    approxDeg = drParam(2);
                end
            end
            if size(highDimData, 2) < preProcessDim
                preProcessDim = size(highDimData, 2);
            end
            
            currPath = pwd;
            fastTsnePath = which('fast_tsne');
            
            cd(fastTsnePath(1:end-11));
            if verbosity
                lowDimData = fast_tsne(highDimData, preProcessDim, perpl, approxDeg);
            else
                [~, lowDimData] = evalc('fast_tsne(highDimData, preProcessDim, perpl, approxDeg);');
            end
            cd(currPath);
            
            clear currPath fastTsnePath;
            drMapping.perpl         = perpl;
            drMapping.preProcessDim = preProcessDim;
            drMapping.approxDeg     = approxDeg;
            
        case {'fisher tsne', 'fisher t-sne'}
            perpl = 30;
            if ~isnan(drParam)
                perpl = drParam;
            end
            
            if verbosity
                lowDimData = tsne_d(dis_Fisher, [], 2, perpl);
            else
                [~, lowDimData] = evalc('tsne_d(dis_Fisher, [], 2, perpl);');
            end
            
            drMapping.perpl = perpl;

        case {'isomap'}
            nNeig = 12;
            if ~isnan(drParam)
                nNeig = drParam;
            end
            lowDimData  = zeros(size(highDimData,1), 2);
            [tmp, mapR] = isomap(highDimData, 2, nNeig);
            lowDimData(1:size(tmp,1),1:size(tmp,2)) = tmp;
            
            drMapping.mapR = mapR;
            clear tmp mapR;
            
            % sort items back (due to unconnected elements, the order
            % doesnt match the order of the label and highDimData)
            nDataWhole = size(highDimData, 1);
            reorderMap = zeros(1, nDataWhole);
            for d=1:nDataWhole
                pos = find(d == drMapping.mapR.conn_comp);
                if isempty(pos)
                    reorderMap(d) = nDataWhole;
                else
                    reorderMap(d) = pos;
                end
            end
            lowDimData = lowDimData(reorderMap, :);
            
        case {'fisher isomap'}
            nNeig = 12;
            if ~isnan(drParam)
                nNeig = drParam;
            end
            lowDimData  = zeros(size(dis_Fisher,1), 2);
            [tmp, mapR] = isomap_d(dis_Fisher, 2, nNeig);
            lowDimData(1:size(tmp,1),1:size(tmp,2)) = tmp;
            
            drMapping.mapR = mapR;
            clear tmp mapR;
            
            % sort items back (due to unconnected elements, the order
            % doesnt match the order of the label and highDimData)
            nDataWhole = size(highDimData, 1);
            reorderMap = zeros(1, nDataWhole);
            for d=1:nDataWhole
                pos = find(d == drMapping.mapR.conn_comp);
                if isempty(pos)
                    reorderMap(d) = nDataWhole;
                else
                    reorderMap(d) = pos;
                end
            end
            lowDimData = lowDimData(reorderMap, :);
            
        case 'mvu'
            nNeig = 12;
            if ~isnan(drParam)
                nNeig = drParam;
            end
            %[tmp, mapR{i}] = fastmvu(X, 2, knn);
            lowDimData  = zeros(size(highDimData,1), 2);
            DD          = dissimilar(highDimData', inf, nNeig);
            [ymvu,mapR] = fastmvu(DD,2,'leigsdim',2,'eta',1e-02,'maxiter',50);
            tmp = mapR.ymvuLap(1:2,:)';
            lowDimData(1:size(tmp,1),1:size(tmp,2)) = tmp;
            
            drMapping.mapR = mapR;
            clear tmp mapR;
            
        case {'fisher mvu'}
            nNeig = 12;
            if ~isnan(drParam)
                nNeig = drParam;
            end
            lowDimData  = zeros(size(dis_Fisher,1), 2);
            DD=dissimilar_d(dis_Fisher, inf, nNeig);
            [ymvuF,mapR]=fastmvu(DD,2,'leigsdim',20,'eta',1e-02,'maxiter',500);
            tmp = mapR.ymvuLap(1:2,:)';
            lowDimData(1:size(tmp,1),1:size(tmp,2)) = tmp;
            
            drMapping.mapR = mapR;
            clear tmp mapR;
            
        case {'som'}                                 % requires the SOM toolbox
            % use the SOM toolbox implementation
            gridSize = [40 40];
            trainLen = 120;
            if ~isnan(drParam(1))
                gridSize = drParam(1:2);
                if length(drParam) > 2
                    trainLen = drParam(3);
                end
            end
            lat = 'rect';

            if ~isfield(drMapping, 'sMap')
                % initialize and train the SOM
                sMap = som_lininit(highDimData, 'msize', gridSize, 'lattice', lat);
                sMap = som_seqtrain(sMap, highDimData, 'trainlen', trainLen);
            else
                sMap = drMapping.sMap;
            end

            if strcmp(sMap.topol.lattice, 'rect')
                % increase the resolution by neuron interpolation
                factor = 2;
                if prod(sMap.topol.msize) > 10e3
                    factor = 0;
                end
                sMap = interpSOM(sMap, factor);
            end
            ldGrid = som_vis_coords(sMap.topol.lattice,sMap.topol.msize);

            % for each data point, find the closest prototype
            lowDimData = zeros(size(highDimData,1),2);
            for i=1:size(highDimData,1)
                [~,ind] = min(sum(bsxfun(@minus, highDimData(i,:), sMap.codebook).^2,2));
                lowDimData(i,:) = ldGrid(ind,:);
            end

            drMapping.sMap   = sMap;

        case {'fisher som', 'fisher rsom'}           % requires the SOM toolbox
            % use the SOM toolbox implementation
            gridSize = [40 40];
            if ~isnan(drParam(1))
                gridSize = drParam(1:2);
            end
            lat = 'rect';

            if ~isfield(drMapping, 'sMap')
                % initialize and train the SOM
                sMap = rsom_lininit(dis_Fisher,  gridSize, 'lattice', lat);
                sMap = rsom_batchtrain(sMap, dis_Fisher);

                % calc neuron positions
                if ~isempty(protsName)
                    nSkip = size(model.(protsName), 1) + 1;
                    sMap.codebook = sMap.cCodebook * highDimData(nSkip:end,:);
                else
                    sMap.codebook = sMap.cCodebook * highDimData;
                end
            else
                sMap = drMapping.sMap;
            end

            if strcmp(sMap.topol.lattice, 'rect')
                % increase the resolution by neuron interpolation
                factor = 3;
                if prod(sMap.topol.msize) > 10e3
                    factor = 0;
                end
                sMap = interpSOM(sMap, factor);
            end
            ldGrid = som_vis_coords(sMap.topol.lattice,sMap.topol.msize);


            % for each data point, find the closest prototype using Fisher Dist
%             lowDimData = zeros(size(highDimData,1),2);
%             if ~exist('idxFisher', 'var')
%                 [~,~, idxFisher{1}, idxFisher{2}] = subSample(highDimData, labels, nDataForParzen);
%             end
%             % calculate the distance to a downsampled part of the neurons
%             [~,ind] = min(compute_fisher_distance_parzen(...
%                 highDimData(idxFisher{1},:), labels(idxFisher{1}), highDimData, ...
%                 sMap.codebook(1:10:end,:)), [], 2);
%             % caculate the distance to the 100 surrounding neurons of each winner
%             ind = (ind - 1)*10 + 1;
%             positions = reshape(1:prod(sMap.topol.msize), ...
%                 sMap.topol.msize(1), sMap.topol.msize(2));
%             for i=1:length(ind)
%                 % calc the dist between each point in hdData and 100 (~81) neurons
%                 [neuronPosx, neuronPosy] = find(positions == ind(i));
%                 % create array with size 9 x 9
%                 tmp = positions(max(neuronPosx-4,1):min(neuronPosx+4, size(positions, 1)), ...
%                     max(neuronPosy-4,1): min(neuronPosy+4, size(positions, 2)));
%                 addressAr = reshape(tmp, numel(tmp), 1);
%                 [~, ind2] = min(compute_fisher_distance_parzen(...
%                     highDimData(idxFisher{1},:), labels(idxFisher{1}), ...
%                     sMap.codebook(addressAr, :), highDimData(i,:)));
%                 lowDimData(i,:) = ldGrid(addressAr(ind2), :);
%             end

            % for each data point, find the closest prototype using eucl Dist
            lowDimData = zeros(size(highDimData,1),2);
            for i=1:size(highDimData,1)
                [~,ind] = min(sum(bsxfun(@minus, highDimData(i,:), sMap.codebook).^2,2));
                lowDimData(i,:) = ldGrid(ind,:);
            end

            drMapping.sMap   = sMap;
            
        case {'gtm'}
            mode  = 'GTM';
            nX    = 30; % 10
            nF    = 6;  % 6
            l2Reg = 40; % 0.2
            
            if ~isnan(drParam(1))
                nX = drParam(1);
                if length(drParam) > 1
                    nF = drParam(2);
                end
                if length(drParam) > 2
                    l2Reg = drParam(3);
                end
            end
            
            Dis = pdist2(highDimData,highDimData);
            
            [X_prot, FI, W, b, MU, sigma] = rel_gtm_run(Dis, labels, mode, ...
                'nX', nX, 'nF', nF, 'l2Reg', l2Reg);
            
            % interpolate the GTM
            X_prot     = gtm_rctg(200, 200);
            FI         = gtm_gbf(MU, sigma, X_prot);
            lowDimData = rel_gtm_pmn(Dis, X_prot, FI, W, b);
            
%             plotData(lowDimData, labels); % debug
%             plot(X_prot(:,1), X_prot(:,2), 'ko');
%             plot(MU(:,1), MU(:,2), 'go');
            
            drMapping.highDimData = highDimData;
            drMapping.X_prot = X_prot;
            drMapping.FI     = FI;
            drMapping.W      = W;
            drMapping.b      = b;
            
            % the centers and widths of the basis functions
            drMapping.MU     = MU;
            drMapping.sigma  = sigma;
            
        case {'fisher gtm', 'fisher rgtm'}
            mode  = 'GTM';
            nX    = 10;
            nF    = 6;
            l2Reg = 0.2;
            
            if ~isnan(drParam(1))
                nX = drParam(1);
                if length(drParam) > 1
                    nF = drParam(2);
                end
                if length(drParam) > 2
                    l2Reg = drParam(3);
                end
            end
            
            [X_prot, FI, W, b, MU, sigma] = rel_gtm_run(dis_Fisher, labels, mode, ...
                'nX', nX, 'nF', nF, 'l2Reg', l2Reg);
            
            % interpolate the GTM
            X_prot     = gtm_rctg(200, 200);
            FI         = gtm_gbf(MU, sigma, X_prot);
            lowDimData = rel_gtm_pmn(dis_Fisher, X_prot, FI, W, b);
            
            drMapping.highDimData = highDimData;
            drMapping.X_prot = X_prot;
            drMapping.FI     = FI;
            drMapping.W      = W;
            drMapping.b      = b;
            
            % the centers and widths of the basis functions
            drMapping.MU     = MU;
            drMapping.sigma  = sigma;

        otherwise
            error('Unknown Dimensionality Reduction method.');
    end
    drMapping.dimRed = dimRed;
    
    % do mean centering for all except PCA and GTM
    if ~(strcmpi(drMapping.dimRed, 'pca') || strcmpi(drMapping.dimRed, 'nodr') ...
            || strcmpi(drMapping.dimRed(end-2:end), 'gtm'))
        meanLd      = mean(lowDimData);
        lowDimData  = bsxfun(@minus, lowDimData, meanLd);
        
        % save for the calculation of the inverse for SOM
        drMapping.meanLd  = meanLd;
    end
    
    % rotate the projected points to be aligned along the PCA components
    if ~(strcmpi(drMapping.dimRed, 'pca') || ...
            strcmpi(drMapping.dimRed(end-2:end), 'som') || ...
            strcmpi(drMapping.dimRed(end-2:end), 'gtm') || ...
            strcmpi(drMapping.dimRed, 'nodr'))
        [eigVecs,~] = eig(cov(lowDimData));
        lowDimData  = lowDimData * eigVecs;

        drMapping.eigVecs = eigVecs;
    end
    
    if isempty(lowDimD4Quality)
        lowDimD4Quality = lowDimData(1:size(highDimD4Quality,1),:);
        lowDimData      = lowDimData(size(highDimD4Quality,1)+1:end,:);
        highDimData     = highDimData(size(highDimD4Quality,1)+1:end,:);
    end
    if ~isempty(protsName)
        nProts      = size(model.(protsName),1);
        lowDimProts = lowDimData(1:nProts,:);
        lowDimData  = lowDimData(nProts+1:end,:);
        highDimData = highDimData(nProts+1:end,:);
    end
end

%% calculate the inverse mapping and visualize the classifier
if ~dr_only
    % if labels for the quality data are provided, plot these as well
    if ~isempty(labelsD4Quality)
        data2plot   = [data2plot; lowDimD4Quality];
        labels2plot = [labels2plot; labelsD4Quality];
        if showVisErrors > 0
            showVisErrors = 2;
        end
    end
    
    [alphaHD, betaHD, lowDimLandmarks, ~,imRes,~,~, classSim] = showClassifCurrentVis(...
              classifPtr, model, lowDimData, highDimData, labels, ...
              'scaleB', scaleB, 'etaAlpha', etaAlpha, ...
              'intensityScale', intensityScale, 'plotIntermResult', plotIntermResult, ...
              'data2plot', data2plot, 'labels2plot', labels2plot, ...
              'estimateQualy', 1, 'lowDimD4Quality', lowDimD4Quality, ...
              'highDimD4Quality', highDimD4Quality, 'showVisErrors', 1, ...
              'alphaHD', alphaHD, 'beta', betaHD, ...
              'lowDimLandmarks', lowDimLandmarks, 'highDimLandmarks', highDimLandmarks, ...
              'lowDimProts', lowDimProts, 'protLabels', protLabels, ...
              'estimateQualy', estimateQualy, 'showVisErrors', showVisErrors, ...
              'J', J, 'drMapping', drMapping, 'verbosity', verbosity, ...
              'nLandmarks', nLandmarks, 'data2plotIdx', data2plotIdx);
end
                  
if nargout >= 1
    % save the return values
    invMap.alphaHD         = alphaHD;
    invMap.betaHD          = betaHD;
    invMap.lowDimLandmarks = lowDimLandmarks;
    if ~isnan(J(1,1,1))
        invMap.J           = J;
    end
    if ~isnan(dis_Fisher)
        invMap.dis_Fisher  = dis_Fisher;
    end
    if ~isnan(lowDimProts)
        invMap.lowDimProts = lowDimProts;
    end
    invMap.lowDimData      = lowDimData;
    if ~isempty(lowDimD4Quality)
        invMap.lowDimD4Quality = lowDimD4Quality;
    end
    if exist('idxFisher', 'var')
        invMap.idxFisher = idxFisher;
    end

    if isfield(drMapping, 'dimRed')
        invMap.drMapping = drMapping;
    end
end

